import json
import logging
import os
import random
import socket

from haigha.connection import Connection as haigha_Connection
from haigha.connections import RabbitConnection
from haigha.message import Message

logger = logging.getLogger('cronq')

class Publisher(object):

    def __init__(self):
        self._connection = connect()
        self._channel = self._connection.channel()

    def publish(self, routing_key, job, run_id):
        cmd = {
            'run_id': str(run_id),
            'job_id': job['id'],
            'cmd': job['command'],
            'name': job['name'],
        }
        print cmd
        self._publish(routing_key, cmd)

    def _publish(self, routing_key, body):
        msg = Message(json.dumps(body), {})
        self._channel.basic.publish(msg, 'cronq', routing_key)


def connect():
    hosts_string = os.getenv('RABBITMQ_HOSTS', None)
    if hosts_string is not None:
        hosts = hosts_string.split(',')
        logger.info('Hosts are: {0}'.format(hosts))
        random.shuffle(hosts)
    else:
        hosts = [os.getenv('RABBITMQ_HOST', 'localhost')]
    user = os.getenv('RABBITMQ_USER', 'guest')
    password = os.getenv('RABBITMQ_PASS', 'guest')
    rabbit_logger = logging.getLogger('amqp-dispatcher.haigha')
    conn = connect_to_hosts(
        RabbitConnection,
        hosts,
        user=user,
        password=password,
        logger=rabbit_logger,
        heartbeat=43200,
    )
    return conn

def connect_to_hosts(connector, hosts, **kwargs):
    for host in hosts:
        logger.info('Trying to connect to host: {0}'.format(host))
        try:
            conn = connector(host=host, **kwargs)
            return conn
        except socket.error:
            logger.info('Error connecting to {0}'.format(host))
    logger.error('Could not connect to any hosts')



import datetime
import logging

from flask import Flask, g, render_template, request, redirect, url_for, flash, abort

from cronq import interval_parser
from backends.mysql import Storage

app = Flask(__name__)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
app.logger.addHandler(stream_handler)


app.secret_key = 'not a secret'

@app.before_request
def create_storage():
    if request.path.startswith('/static/'):
        return
    g.storage = Storage()

@app.after_request
def remove_storage(request):
    if hasattr(g, 'storage'):
        try:
            g.storage.close()
        except Exception as exc:
            print exc
    return request

@app.route('/')
def index():
    jobs = list(g.storage.jobs)
    return render_template('index.html', jobs=jobs)


@app.route('/job/<int:id>', methods=['GET', 'POST'])
def job(id):
    if request.method == 'POST' and request.form.get('run_now') is not None:
        g.storage.run_job_now(id)
        flash('Job submitted at {0}'.format(datetime.datetime.utcnow()))
        return redirect(url_for('job', id=id))

    job_doc = g.storage.get_job(id)
    events = g.storage.last_events_for_job(id, 10)
    return render_template('job.html', job=job_doc, events=events)

@app.route('/run/<string:id>')
def run_id(id):
    events = list(g.storage.events_for_run_id(id))
    job_id = events[0]['job_id']
    job = g.storage.get_job(job_id)
    return render_template('run_id.html', events=events, job=job)

@app.route('/failures')
def failures():
    failure_events = list(g.storage.failures())
    names = { job['id']: job['name'] for job in g.storage.jobs }
    for event in failure_events:
        event['job_name'] = names[event['job_id']]
    return render_template('failures.html', events=failure_events)

@app.route('/api/category/<string:name>', methods=['PUT', 'POST'])
def category(name):
    data = request.json
    existing_jobs = g.storage.jobs_for_category(name=name)
    category_id = g.storage.category_id_for_name(name)
    job_lookup = {}

    if not validate_unique_job_names(data.get('jobs', [])):
        abort(400)

    for job in existing_jobs:
        job_lookup[job['name']] = job

    for job in data.get('jobs', []):
        name = job['name']
        next_run, duration = interval_parser.next_run_and_duration_from_8601(job['schedule'])
        existing_job = job_lookup.get(name, {})
        new_id = existing_job.get('id')
        new_next_run = next_run
        new_interval = duration.total_seconds()
        command = job['command']
        g.storage.add_job(
            name,
            new_interval,
            command,
            next_run,
            new_id,
            category_id,
            routing_key=job.get('routing_key')
        )
        if existing_job:
            del job_lookup[name]

    remove_jobs(g.storage, job_lookup.itervalues())

    return '{"status": "success"}'


def remove_jobs(storage, jobs):
    for job in jobs:
        g.storage.remove_job(job['id'])


def validate_unique_job_names(jobs):
    job_names = [job['name'] for job in jobs]
    return len(job_names) == len(set(job_names))


if __name__ == "__main__":
    app.run(debug=True)



import datetime
import aniso8601

def next_run_and_duration_from_8601(interval):
    gen = aniso8601.parse_repeating_interval(interval)
    start = next(gen)
    second = next(gen)
    duration = second - start
    now = datetime.datetime.utcnow()
    diff_from_now = now - start
    number_of_iterations = number_of_iterations_in_time(duration, diff_from_now)
    next_run = apply_delta_to_datetime(duration, start, number_of_iterations +1)
    return next_run, duration


def number_of_iterations_in_time(duration, time):
    """Return the number of times full duration is in time"""
    return int(time.total_seconds() / duration.total_seconds())


def apply_delta_to_datetime(delta, dt, times):
    full_td = delta * times
    final_dt = dt + full_td
    return final_dt



import time

class Injector(object):

    def __init__(self, storage):
        self.storage = storage
        storage.bootstrap()

    def run(self):
        while True:
            self.storage.inject()
            time.sleep(1)

def main():
    from cronq.backends.mysql import Storage
    from cronq.queue_connection import Publisher
    injector = Injector(Storage(Publisher()))
    injector.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def run_setup():
    setup(
        name='cronq',
        version='0.0.27',
        description='A Cron-like system for running tasks',
        keywords = 'cron amqp',
        url='http://github.com/philipcristiano/cronq',
        author='Philip Cristiano',
        author_email='philipcristiano@gmail.com',
        license='BSD',
        packages=['cronq', 'cronq.backends'],
        install_requires=[
            'aniso8601==0.70',
            'flask',
            'haigha',
            'sqlalchemy',
            'mysql-connector-python',
            'python-dateutil',
        ],
        test_suite='tests',
        long_description=read('README.md'),
        include_package_data=True,
        zip_safe=True,
        classifiers=[
        ],
        entry_points="""
        [console_scripts]
           cronq-runner=cronq.runner:main
           cronq-injector=cronq.injector:main
           cronq-results=cronq.result_aggregator:main
        """,
    )

if __name__ == '__main__':
    run_setup()



from uuid import uuid4, UUID
import datetime
import os
import socket

import sqlalchemy
from sqlalchemy import (
    or_,
    Binary,
    Column,
    CHAR,
    DateTime,
    Integer,
    Interval,
    String,
    Time,
    Text,
    UniqueConstraint
)
from sqlalchemy.sql.expression import desc
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.orm import sessionmaker

Base = declarative_base()



class Storage(object):

    def __init__(self, publisher=None):
        self.publisher = publisher

        self._engine = new_engine()
        self._maker = sessionmaker(bind=self._engine)
        self.session = self._new_session()

    def _new_session(self):
        return self._maker()

    def bootstrap(self):
        models = [
            Job,
            Event,
            Category,
        ]
        for model in models:
            try:
                print 'Try model', model
                model.__table__.create(self._engine)
            except (OperationalError, ProgrammingError) as exc:
                print exc
                pass

    def close(self):
        self.session.close()
        self._engine.dispose()

    def add_job(self, name, interval_seconds, command, next_run, id=None, category_id=None, routing_key=None):
        if routing_key is None:
            routing_key = 'default'
        job = Job()
        job.id = id
        job.name = name
        job.next_run = next_run
        job.interval = datetime.timedelta(seconds=interval_seconds)
        job.command = command
        job.routing_key = routing_key
        job.category_id = category_id
        self.session.merge(job)
        self.session.commit()

    def remove_job(self, job_id):
        job = self.session.query(Job).filter_by(id=job_id).first()
        if job:
            self.session.delete(job)
            self.session.commit()

    def add_event(self, job_id, datetime, run_id, type, host, return_code):
        event = Event()
        event.job_id = job_id
        event.datetime = datetime
        event.run_id = run_id
        event.type = type
        event.host = host
        event.return_code = return_code
        self.session.add(event)
        self.session.commit()

    @property
    def jobs(self):
        session = self.session
        for job in session.query(Job):
            yield {
                'id': job.id,
                'name': job.name,
                'next_run': job.next_run,
                'interval': job.interval,
            }

    def get_job(self, id):
        job = self.session.query(Job).filter_by(id=id).first()
        return self._job_doc_for_model(job)

    def _job_doc_for_model(self, job):
        return {
            'id': job.id,
            'name': job.name,
            'next_run': job.next_run,
            'interval': job.interval,
            'command': job.command,
        }


    def jobs_for_category(self, id=None, name=None):
        if id and name:
            assert "Don't pass both id and name"
        if name:
            id = self.category_id_for_name(name)
        for job in self.session.query(Job).filter_by(category_id=id):
            yield self._job_doc_for_model(job)

    def category_id_for_name(self, name):
        category = self.session.query(Category).filter_by(name=name).first()
        if category is None:
            category = Category(name=name)
            self.session.add(category)
            self.session.commit()
        return category.id


    def last_events_for_job(self, job_id, number):
        events = self.session.query(Event).filter_by(job_id=job_id).order_by(desc(Event.id)).limit(number)
        return event_models_to_docs(events)

    def events_for_run_id(self, run_id):
        events = self.session.query(Event).filter_by(run_id=run_id).order_by(Event.id)
        return event_models_to_docs(events)

    def failures(self):
        events = self.session.query(Event)\
            .filter_by(type='finished')\
            .filter(Event.return_code!=0)\
            .order_by(desc(Event.id)).limit(50)
        return event_models_to_docs(events)

    def run_job_now(self, id):
        event = self.session.query(Job).filter_by(id=id).first()
        event.run_now = True
        self.session.commit()

    def inject(self):
        """Get a message from storage and injects it into the job stream"""
        print 'Trying to inject', datetime.datetime.utcnow()
        while self.get_unpublished_task() is not None:
            pass

    def get_unpublished_task(self):
        session = self._new_session()
        to_run = or_(
            Job.next_run < datetime.datetime.utcnow(),
            Job.run_now == True,
        )

        job = session.query(Job).filter(to_run).first()
        if job is None:
            session.close()
            return
        print 'Found a job:', job.name, job.next_run
        to_run_at = job.next_run
        while job.next_run < datetime.datetime.utcnow():
            print 'Adding time!'
            job.next_run += job.interval
        print job.next_run
        job_doc = {
            'name': job.name,
            'command': unicode(job.command),
            'id': job.id,
        }
        job.run_now = False
        me = '{0}.{1}'.format(socket.gethostname(), os.getpid())
        job.locked_by = me
        try:
            session.commit()
            self.publisher.publish(job.routing_key, job_doc, uuid4().hex)
        except Exception as exc:
            session.rollback()
            raise
        session.close()
        return True

def event_models_to_docs(events):
    for event in events:
        yield {
            'id': event.id,
            'datetime': event.datetime,
            'run_id': event.run_id,
            'host': event.host,
            'return_code': event.return_code,
            'type': event.type,
            'job_id': event.job_id,
        }


def new_engine():
    dsn = os.getenv('CRONQ_MYSQL', 'mysql+mysqlconnector://root@localhost/cronq')
    return create_engine(dsn, isolation_level='SERIALIZABLE')


class Job(Base):

    __tablename__ = 'jobs'
    __table_args__ = (UniqueConstraint('category_id', 'name'),{'mysql_engine':'InnoDB'})

    id = Column(Integer, primary_key=True)
    name = Column(CHAR(255))
    interval = Column(Interval)
    next_run = Column(DateTime(), default=datetime.datetime.utcnow)
    routing_key = Column(CHAR(32), default='default')
    command = Column(Text())
    run_now = Column(Integer)
    locked_by = Column(CHAR(64))
    category_id = Column(Integer)


class Event(Base):

    __tablename__ = 'events'
    __table_args__ = {'mysql_engine':'InnoDB'}

    id = Column(Integer, primary_key=True)
    job_id = Column(Integer)
    datetime = Column(DateTime)
    run_id = Column(CHAR(32))
    type = Column(CHAR(32))
    host= Column(CHAR(255))
    return_code = Column(Integer)

class Category(Base):

    __tablename__ = 'categories'
    __table_args__ = {'mysql_engine':'InnoDB'}

    id = Column(Integer, primary_key=True)
    name = Column(CHAR(255), unique=True)
